/**
 * @file
 * @brief
 *
 * @author  Anton Kozlov
 * @date    10.11.2014
 */

#include <assert.h>
#include <stdbool.h>
#include <util/array.h>

#include <stm32f4xx_gpio.h>
#include <stm32f4xx_rcc.h>

#include "libleddrv_ll.h"

#define LEDS_PER_LINE 8
#define LINES_N 10
static_assert(LINES_N * LEDS_PER_LINE == LEDDRV_LED_N);

struct leddrv_pin_desc {
	GPIO_TypeDef *gpio; /**< gpio port  */
	unsigned int pin; /**< pin mask */
};

#define LEDBLOCK_MAJOR 2
#define LEDBLOCK_MINOR 0

#if LEDBLOCK_MAJOR == 0
static const struct leddrv_pin_desc leddrv_clk =
	{ .gpio = GPIOE, .pin = GPIO_Pin_7 };

static const struct leddrv_pin_desc leddrv_datas[] = {
	/* prototype register numeration */
	{ .gpio = GPIOB, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOB, .pin = GPIO_Pin_15 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_15 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_13 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_11 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_12 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_10 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_9 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_8 },
};
#endif

#if LEDBLOCK_MAJOR == 1 && LEDBLOCK_MINOR == 0
static const struct leddrv_pin_desc leddrv_clk =
	{ .gpio = GPIOE, .pin = GPIO_Pin_7 };

static const struct leddrv_pin_desc leddrv_datas[] = {
	/* v1 register numeration, forward, wrong */
	{ .gpio = GPIOE, .pin = GPIO_Pin_12 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_13 },
	{ .gpio = GPIOB, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOB, .pin = GPIO_Pin_15 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_15 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_11 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_10 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_9 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_8 },
};
#endif

#if LEDBLOCK_MAJOR == 1 && LEDBLOCK_MINOR == 1
static const struct leddrv_pin_desc leddrv_clk =
	{ .gpio = GPIOE, .pin = GPIO_Pin_8 };

static const struct leddrv_pin_desc leddrv_datas[] = {
	/* prototype register numeration */
	{ .gpio = GPIOE, .pin = GPIO_Pin_11 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_13 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOB, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_8 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_15 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_12 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_9 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_10 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_7 },
};
#endif

#if LEDBLOCK_MAJOR == 2
static const struct leddrv_pin_desc leddrv_shf_clk =
	{ .gpio = GPIOE, .pin = GPIO_Pin_8 };
static const struct leddrv_pin_desc leddrv_str_clk =
	{ .gpio = GPIOD, .pin = GPIO_Pin_9 };
static const struct leddrv_pin_desc leddrv_datas[] = {
	{ .gpio = GPIOE, .pin = GPIO_Pin_11 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_13 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOB, .pin = GPIO_Pin_14 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_8 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_15 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_12 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_9 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_10 },
	{ .gpio = GPIOE, .pin = GPIO_Pin_7 },
};

static const struct leddrv_pin_desc leddrv_alarms[] = {
	{ .gpio = GPIOD, .pin = GPIO_Pin_3 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_4 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_1 },
	{ .gpio = GPIOC, .pin = GPIO_Pin_12 },
	{ .gpio = GPIOA, .pin = GPIO_Pin_10 },
	{ .gpio = GPIOA, .pin = GPIO_Pin_8 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_2 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_7 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_5 },
	{ .gpio = GPIOD, .pin = GPIO_Pin_6 },
};
#endif

static_assert(ARRAY_SIZE(leddrv_datas) == LINES_N);
static_assert(ARRAY_SIZE(leddrv_alarms) == LINES_N);

int leddrv_ll_error(int n) {
	int line = n / LEDS_PER_LINE;
	return !(GPIO_ReadInputData(leddrv_alarms[line].gpio) & leddrv_alarms[line].pin);
}

static inline void leddrv_pin_init(const struct leddrv_pin_desc *pd, unsigned int mode) {
	GPIO_InitTypeDef GPIO_InitStructure = {
		.GPIO_Speed = GPIO_Speed_50MHz,
		.GPIO_Mode  = mode,
		.GPIO_OType = GPIO_OType_PP,
		.GPIO_PuPd  = GPIO_PuPd_NOPULL,
	};
	GPIO_InitStructure.GPIO_Pin = pd->pin;
	GPIO_Init(pd->gpio, &GPIO_InitStructure);
}

void leddrv_ll_init(void) {
	/* Enable GPIOs clocks */
	RCC_AHB1PeriphClockCmd(
			RCC_AHB1Periph_GPIOA |
			RCC_AHB1Periph_GPIOB |
			RCC_AHB1Periph_GPIOC |
			RCC_AHB1Periph_GPIOD |
			RCC_AHB1Periph_GPIOE,
			ENABLE);

	leddrv_pin_init(&leddrv_shf_clk, GPIO_Mode_OUT);
	leddrv_pin_init(&leddrv_str_clk, GPIO_Mode_OUT);
	for (int i = 0; i < LINES_N; ++i) {
		leddrv_pin_init(&leddrv_datas[i], GPIO_Mode_OUT);
	}

	for (int i = 0; i < LINES_N; ++i) {
		leddrv_pin_init(&leddrv_alarms[i], GPIO_Mode_IN);
	}
}

#if LEDBLOCK_MAJOR == 1
static void leddrv_ll_shift_out(int led_in_line, unsigned char leds_state[LEDDRV_LED_N]) {

	/* set clock low */
	GPIO_ResetBits(leddrv_clk.gpio, leddrv_clk.pin);

	/* set data in for all `led_in_line'-th led */
	for (int line = 0; line < LINES_N; ++line) {
		const struct leddrv_pin_desc *line_desk = &leddrv_datas[line];
		const bool led_state = leds_state[led_in_line + line * LEDS_PER_LINE ];

		if (led_state) {
			/* set data high */
			GPIO_SetBits(line_desk->gpio, line_desk->pin);
		} else {
			/* set data low */
			GPIO_ResetBits(line_desk->gpio, line_desk->pin);
		}
	}

	/* make clock high, shift data to regsiter */
	GPIO_SetBits(leddrv_clk.gpio, leddrv_clk.pin);
}


void leddrv_ll_update(unsigned char leds_state[LEDDRV_LED_N]) {

	/* shift-register connected as
	 * shift-out-0 -> led-4
	 * shift-out-1 -> led-5
	 * shift-out-2 -> led-6
	 * shift-out-3 -> led-7
	 * shift-out-4 -> led-3
	 * shift-out-5 -> led-2
	 * shift-out-6 -> led-1
	 * shift-out-7 -> led-0
	 * so have to shift first 4 in asceding, then next 4 in desceding
	 */

	for (int led_in_line = 0; led_in_line < 4; ++led_in_line) {
		leddrv_ll_shift_out(led_in_line, leds_state);
	}

	for (int led_in_line = LEDS_PER_LINE - 1; led_in_line >= 4; --led_in_line) {
		leddrv_ll_shift_out(led_in_line, leds_state);
	}

}
#endif

#if LEDBLOCK_MAJOR == 2
void leddrv_ll_update(unsigned char leds_state[LEDDRV_LED_N]) {

	GPIO_ResetBits(leddrv_str_clk.gpio, leddrv_str_clk.pin);

	for (int i_led = LEDS_PER_LINE - 1; i_led >= 0; --i_led) {

		GPIO_ResetBits(leddrv_shf_clk.gpio, leddrv_shf_clk.pin);

		for (int i_line = 0; i_line < LINES_N; ++i_line) {
			const struct leddrv_pin_desc *line_desk = &leddrv_datas[i_line];
			const bool led_state = leds_state[i_line * LEDS_PER_LINE + i_led];

			if (led_state) {
				GPIO_SetBits(line_desk->gpio, line_desk->pin);
			} else {
				GPIO_ResetBits(line_desk->gpio, line_desk->pin);
			}
		}

		GPIO_SetBits(leddrv_shf_clk.gpio, leddrv_shf_clk.pin);
	}

	GPIO_SetBits(leddrv_str_clk.gpio, leddrv_str_clk.pin);
}
#endif
